import sys
import argparse
import ply.yacc
import tempfile
import pathlib

from ..parse import Parser, LexError, ParseError


def main(args: list[str] = sys.argv[1:]) -> None:
    argparser: argparse.ArgumentParser = argparse.ArgumentParser(
        prog="typoc",
        usage="typoc [-O] [-o <file>] <input>",
        description="typoc TEVM compiler",
        add_help=True,
    )

    argparser.add_argument(
        "-O",
        "--optimize",
        dest="optimize",
        action="store_true",
        help="Enable optimizations",
    )

    argparser.add_argument(
        dest="<input>",
        type=pathlib.Path,
        help="File to compile",
    )

    argparser.add_argument(
        "-o",
        dest="output",
        metavar="<file>",
        type=pathlib.Path,
        help="Write output to <file>",
    )

    args: argparse.Namespace = argparser.parse_args(args)

    input: pathlib.Path = getattr(args, "<input>")
    output: pathlib.Path = getattr(args, "output")
    optimize: bool = getattr(args, "optimize")

    if output is None:
        output = input.with_suffix(".bin")

    if not input.is_file():
        print(f"error: {input} is not a file")

        exit(1)

    if output.is_dir():
        print(f"error: {output} is a directory")

    tree: list = []

    with tempfile.TemporaryDirectory() as cache_path:
        try:
            parser: Parser = Parser(
                input.read_text(), optimize, pathlib.Path(cache_path)
            )

            try:
                tree = parser.parse()

            except LexError as lex_error:
                print(f"error: {lex_error.message}")

                try:
                    print(f"| {parser.source.splitlines()[lex_error.token.lineno - 1]}")

                except IndexError:
                    pass

                try:
                    print(f"| {parser.source.splitlines()[lex_error.token.lineno]}")

                except IndexError:
                    pass

                try:
                    print(f"| {parser.source.splitlines()[lex_error.token.lineno + 1]}")

                except IndexError:
                    pass

                print(
                    f"{input}:{lex_error.token.lineno}:{parser.find_column(lex_error.token)}"
                )

                exit(1)

            except ParseError as parse_error:
                print(f"error: {parse_error.message}")

                try:
                    print(
                        f"| {parser.source.splitlines()[parse_error.prod.lineno - 1]}"
                    )

                except IndexError:
                    pass

                try:
                    print(f"| {parser.source.splitlines()[parse_error.prod.lineno]}")

                except IndexError:
                    pass

                try:
                    print(
                        f"| {parser.source.splitlines()[parse_error.prod.lineno + 1]}"
                    )

                except IndexError:
                    pass

                print(
                    f"{input}:{parse_error.prod.lineno}:{parser.find_column(parse_error.prod)}"
                )

                exit(1)

        except ply.yacc.YaccError as yacc_error:
            print(f"error: {yacc_error}".lower())

            exit(1)

    code: str = f"{tree}"
    res: bytes = code.encode()

    output.write_bytes(res)
