import ply.lex
import ply.yacc
import pathlib
import cgen


class Parser:
    start: str = "root"
    reserved: dict[str, str] = {
        "import": "IMPORT",
        "static": "STATIC",
        "class": "CLASS",
        "struct": "STRUCT",
        "new": "NEW",
        "delete": "DELETE",
        "this": "THIS",
        "super": "SUPER",
        "if": "IF",
        "else": "ELSE",
        "while": "WHILE",
        "for": "FOR",
        "with": "WITH",
        "return": "RETURN",
        "break": "BREAK",
        "continue": "CONTINUE",
        "switch": "SWITCH",
        "let": "LET",
        "const": "CONST",
        "true": "TRUE",
        "false": "FALSE",
        "null": "NULL",
        "public": "PUBLIC",
        "private": "PRIVATE",
    }

    tokens: tuple[str] = (
        "STRING",
        "INTEGER",
        "FLOAT",
        "CHARACTER",
        "IDENTIFIER",
        "ARROW",
        "DOUBLE_COLON",
        "LEFT_SHIFT",
        "RIGHT_SHIFT",
        "PLUS",
        "MINUS",
        "MULTIPLY",
        "DIVIDE",
        "INCREMENT",
        "DECREMENT",
        "REMAIN",
        "LESS",
        "LESS_EQUAL",
        "GREATER",
        "GREATER_EQUAL",
        "EQUAL",
        "NOT_EQUAL",
        "BIT_AND",
        "BIT_OR",
        "BIT_XOR",
        "AND",
        "OR",
    ) + tuple(reserved.values())

    literals: str = "()[]{}.,;:=?@!+-~"

    t_ignore: str = " \t\r"
    t_ignore_COMMENT: str = r"(\/\/.*)|(\/\*((.|\n)*)\*\/)"

    t_ARROW: str = r"\-\>"
    t_DOUBLE_COLON: str = r"\:\:"
    t_LEFT_SHIFT: str = r"\<\<"
    t_RIGHT_SHIFT: str = r"\>\>"
    t_PLUS: str = r"\+"
    t_MINUS: str = r"\-"
    t_MULTIPLY: str = r"\*"
    t_DIVIDE: str = r"\/"
    t_INCREMENT: str = r"\+\+"
    t_DECREMENT: str = r"\-\-"
    t_REMAIN: str = r"\%"
    t_LESS: str = r"\<"
    t_LESS_EQUAL: str = r"\<\="
    t_GREATER: str = r"\>"
    t_GREATER_EQUAL: str = r"\>\="
    t_EQUAL: str = r"\=\="
    t_NOT_EQUAL: str = r"\!\="
    t_BIT_AND: str = r"\&"
    t_BIT_OR: str = r"\|"
    t_BIT_XOR: str = r"\^"
    t_AND: str = r"\&\&"
    t_OR: str = r"\|\|"

    precedence: tuple = (
        ("left", "PLUS", "MINUS"),
        ("left", "MULTIPLY", "DIVIDE"),
    )

    __source: str = str()

    __optimize: bool = False
    __cache_path: pathlib.Path = pathlib.Path.cwd() / "__pycache__" / "typoc"

    __lexer: ply.lex.Lexer = None
    __parser: ply.yacc.LRParser = None

    def __init__(
        self,
        source: str,
        optimize: bool = True,
        cache_path: pathlib.Path = pathlib.Path.cwd() / "__pycache__" / "typoc",
    ) -> None:
        self.__source: str = source
        self.__optimize: bool = optimize
        self.__cache_path: pathlib.Path = cache_path

        self.__cache_path.mkdir(parents=True, exist_ok=True)

        self.__lexer: ply.lex.Lexer = ply.lex.lex(
            module=self,
            debug=not self.__optimize,
            optimize=self.__optimize,
            lextab="typoc_lextab",
            outputdir=self.__cache_path,
        )

        self.__parser: ply.yacc.LRParser = ply.yacc.yacc(
            method="LALR",
            debug=not self.__optimize,
            module=self,
            tabmodule="typoc_parsetab",
            start="root",
            check_recursion=True,
            optimize=self.__optimize,
            write_tables=True,
            debugfile=self.__cache_path / "typoc_parsetab.out",
            outputdir=self.__cache_path,
        )

        self.__lexer.input(source)

    @property
    def source(self) -> str:
        return self.__source

    @property
    def optimize(self) -> bool:
        return self.__optimize

    @property
    def cache_path(self) -> pathlib.Path:
        return self.__cache_path

    @property
    def lexer(self) -> ply.lex.Lexer:
        return self.__lexer

    @property
    def parser(self) -> ply.yacc.LRParser:
        return self.__parser

    def find_column(self, token):
        line_start = self.__source.rfind("\n", 0, token.lexpos) + 1

        return (token.lexpos - line_start) + 1

    def t_newline(self, t):
        r"\n+"

        t.lexer.lineno += t.value.count("\n")

    def t_error(self, t):
        raise LexError(t, f"'{t.value[0]}': illegal character")

    def p_error(self, p):
        raise ParseError(p, "illegal syntax")

    def t_STRING(self, t):
        r"([\"]([^\\\"]|\\.|\\\n)*[\"])"

        t.value = t.value[1:-1]

        return t

    def t_FLOAT(self, t):
        r"([0-9]+\.[0-9]+)"

        t.value = float(t.value)

        return t

    def t_INTEGER(self, t):
        r"([0-9]+)"

        t.value = int(t.value)

        return t

    def t_CHARACTER(self, t):
        r"([\']([^\\\']|\\.|\\\n)*[\'])"

        t.value = t.value[1:-1]

        if t.value[0] != "\\" and len(t.value) != 1:
            raise LexError(t, f"'{t.value}': character is too long")

        elif t.value[0] == "\\" and len(t.value) != 2:
            raise LexError(t, f"'{t.value}': character is too long")

        return t

    def t_IDENTIFIER(self, t):
        r"[a-zA-Z_][a-zA-Z0-9_]*"

        t.type = self.reserved.get(t.value, "IDENTIFIER")

        return t

    def p_root(self, p):
        """
        root :
        """

        p[0] = cgen.Collection([])

    def parse(self) -> cgen.Collection:
        return self.__parser.parse(
            self.__source,
            lexer=self.__lexer,
            debug=not self.__optimize,
        )

    def __repr__(self) -> str:
        return f"Parser(source={self.__source!r}, optimize={self.__optimize!r})"

    def __str__(self) -> str:
        return self.__repr__()


class LexError(Exception):
    def __init__(self, token: ply.lex.LexToken, message: str) -> None:
        self.__token: ply.lex.LexToken = token
        self.__message: str = message

        super().__init__(f"{self.__token!r}: {self.__message}")

    @property
    def token(self) -> ply.lex.LexToken:
        return self.__token

    @property
    def message(self) -> str:
        return self.__message


class ParseError(Exception):
    def __init__(self, prod: ply.yacc.YaccProduction, message: str) -> None:
        self.__prod: ply.yacc.YaccProduction = prod
        self.__message: str = message

        super().__init__(f"{self.__prod!r}: {self.__message}")

    @property
    def prod(self) -> ply.yacc.YaccProduction:
        return self.__prod

    @property
    def message(self) -> str:
        return self.__message
